#include <d3d8.h>

#include ".\GSEdit.h"
#include ".\GSMenu.h"
#include ".\LD.h"



CGSEdit::CGSEdit()
{

  m_EditMode        = EM_TILES;
  m_CurTile         = LD::TT_DIRT;
  m_ButtonReleased  = false;

  m_CurObjectType   = LD::OT_PLAYER;
  m_pSelectedObj    = NULL;

  m_EditWorld = "world";
  m_Level.Load( m_EditWorld );

}



CGSEdit::~CGSEdit()
{

}



void CGSEdit::Display( IDirect3DDevice8* pDevice )
{

  theApp.m_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, FALSE );

  theApp.m_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLOROP, D3DTOP_MODULATE );
  theApp.m_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
  theApp.m_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLORARG2, D3DTA_DIFFUSE );

  theApp.m_pd3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAOP, D3DTOP_MODULATE );
  theApp.m_pd3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE );
  theApp.m_pd3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAARG2, D3DTA_DIFFUSE );

  m_Level.Display();

  if ( m_EditMode == EM_TILES )
  {
    int tileX = m_Level.TileXFromScreenPos( theApp.m_MouseX - 320 );
    int tileY = m_Level.TileYFromScreenPos( theApp.m_MouseY - 240 );
    int   sectX = tileX / 5;
    if ( m_Level.m_OffsetX + theApp.m_MouseX - 320 < 0 )
    {
      sectX = ( tileX + 1 ) / 5 - 1;
    }
    int   sectY = tileY / 5;
    if ( m_Level.m_OffsetY + theApp.m_MouseY - 240 < 0 )
    {
      sectY = ( tileY + 1 ) / 5 - 1;
    }
    int     inSectX = tileX - sectX * 5;
    int     inSectY = tileY - sectY * 5;

    theApp.RenderTextureSectionAsTile2d( theApp.m_MouseX - 16, theApp.m_MouseY - 16, theApp.m_TexSec[theApp.m_Tile[m_CurTile]] );

    char    temp[200];
    sprintf( temp, "Tile: %d,%d  Sect: %d,%d  InSect: %d,%d", tileX, tileY, sectX, sectY, inSectX, inSectY );
    theApp.Print( 2, 2, temp );
  }
  else if ( m_EditMode == EM_OBJECTS )
  {
    m_Level.DisplayVisibleBounds();

    if ( m_pSelectedObj )
    {
      RECT    bounds = m_pSelectedObj->Bounds();

      OffsetRect( &bounds, -m_Level.m_OffsetX + 320, -m_Level.m_OffsetY + 240 );

      DWORD   color = 0xffffff00;
      theApp.RenderQuad2d( bounds.left, bounds.top, bounds.right - bounds.left, 1, color );
      theApp.RenderQuad2d( bounds.left, bounds.top, 1, bounds.bottom - bounds.top, color );
      theApp.RenderQuad2d( bounds.right - 1, bounds.top, 1, bounds.bottom - bounds.top, color );
      theApp.RenderQuad2d( bounds.left, bounds.bottom - 1, bounds.right - bounds.left, 1, color );
    }

    GameObject    obj( theApp.m_MouseX + m_Level.m_OffsetX - 320, theApp.m_MouseY + m_Level.m_OffsetY - 240, m_CurObjectType );

    obj.Display( m_Level );

    if ( m_pSelectedObj != NULL )
    {
      char    temp[200];
      sprintf( temp, "ObjID: %d   Pos: %d,%d", m_pSelectedObj->m_ID, theApp.m_MouseX + m_Level.m_OffsetX - 320, theApp.m_MouseY + m_Level.m_OffsetY - 240 );
      theApp.Print( 5, 5, temp );
    }
    char    temp[200];
    sprintf( temp, "Pos: %d,%d", theApp.m_MouseX + m_Level.m_OffsetX - 320, theApp.m_MouseY + m_Level.m_OffsetY - 240 );
    theApp.Print( 200, 5, temp );
  }

}



void CGSEdit::UpdateFrame( const float fElapsedTime )
{

  if ( ( theApp.m_bKeyPressed[VK_F2] )
  &&   ( theApp.m_bKeyReleased[VK_F2] ) )
  {
    theApp.m_bKeyReleased[VK_F2] = false;

    m_Level.Save( m_EditWorld );
  }

}



void CGSEdit::UpdateFixed()
{

  if ( theApp.m_bKeyPressed[VK_LEFT] )
  {
    if ( ( m_EditMode == EM_OBJECTS )
    &&   ( m_pSelectedObj != NULL )
    &&   ( theApp.m_bKeyPressed[VK_SHIFT] ) )
    {
      m_Level.MoveObject( m_pSelectedObj, m_pSelectedObj->m_X - 1, m_pSelectedObj->m_Y );
    }
    else
    {
      m_Level.m_OffsetX -= 8;
    }
  }
  if ( theApp.m_bKeyPressed[VK_RIGHT] )
  {
    if ( ( m_EditMode == EM_OBJECTS )
    &&   ( m_pSelectedObj != NULL )
    &&   ( theApp.m_bKeyPressed[VK_SHIFT] ) )
    {
      m_Level.MoveObject( m_pSelectedObj, m_pSelectedObj->m_X + 1, m_pSelectedObj->m_Y );
    }
    else
    {
      m_Level.m_OffsetX += 8;
    }
  }
  if ( theApp.m_bKeyPressed[VK_UP] )
  {
    if ( ( m_EditMode == EM_OBJECTS )
    &&   ( m_pSelectedObj != NULL )
    &&   ( theApp.m_bKeyPressed[VK_SHIFT] ) )
    {
      m_Level.MoveObject( m_pSelectedObj, m_pSelectedObj->m_X, m_pSelectedObj->m_Y - 1 );
    }
    else
    {
      m_Level.m_OffsetY -= 8;
    }
  }
  if ( theApp.m_bKeyPressed[VK_DOWN] )
  {
    if ( ( m_EditMode == EM_OBJECTS )
    &&   ( m_pSelectedObj != NULL )
    &&   ( theApp.m_bKeyPressed[VK_SHIFT] ) )
    {
      m_Level.MoveObject( m_pSelectedObj, m_pSelectedObj->m_X, m_pSelectedObj->m_Y + 1 );
    }
    else
    {
      m_Level.m_OffsetY += 8;
    }
  }

}



void CGSEdit::OnMouse( int X, int Y, int Buttons )
{

  if ( ( Buttons & 1 ) == 0 )
  {
    m_ButtonReleased = true;
  }
  switch ( m_EditMode )
  {
    case EM_TILES:
      {
        int     tileX = ( m_Level.m_OffsetX + theApp.m_MouseX - 320 ) / 32;
        if ( m_Level.m_OffsetX + theApp.m_MouseX - 320 < 0 )
        {
          tileX--;
        }
        int     tileY = ( m_Level.m_OffsetY + theApp.m_MouseY - 240 ) / 32;
        if ( m_Level.m_OffsetY + theApp.m_MouseY - 240 < 0 )
        {
          tileY--;
        }
        int   sectX = tileX / 5;
        if ( m_Level.m_OffsetX + theApp.m_MouseX - 320 < 0 )
        {
          sectX = ( tileX + 1 ) / 5 - 1;
        }
        int   sectY = tileY / 5;
        if ( m_Level.m_OffsetY + theApp.m_MouseY - 240 < 0 )
        {
          sectY = ( tileY + 1 ) / 5 - 1;
        }
        int     inSectX = tileX - sectX * 5;
        int     inSectY = tileY - sectY * 5;

        if ( Buttons & 1 )
        {
          m_Level.m_Sectors[std::make_pair( sectX, sectY )].Ground[inSectX][inSectY] = m_CurTile;
        }
        if ( Buttons & 2 )
        {
          m_CurTile = m_Level.m_Sectors[std::make_pair( sectX, sectY )].Ground[inSectX][inSectY];
        }
        //theApp.RenderTextureSectionAsTile2d( theApp.m_MouseX - 16, theApp.m_MouseY - 16, theApp.m_TexSec[theApp.m_Tile[m_CurTile]] );
      }
      break;
    case EM_OBJECTS:
      if ( ( Buttons & 1 )
      &&   ( m_ButtonReleased ) )
      {
        m_ButtonReleased = false;

        GameObject*   pObjBelow = m_Level.FindObject( X - 320, Y - 240 );
        if ( pObjBelow != NULL )
        {
          m_pSelectedObj = pObjBelow;
        }
        else
        {
          GameObject*   pObj = new GameObject( theApp.m_MouseX + m_Level.m_OffsetX - 320, theApp.m_MouseY + m_Level.m_OffsetY - 240, m_CurObjectType );

          pObj->m_ID = m_Level.NewID();

          m_Level.AddObject( pObj );
          m_pSelectedObj = pObj;
        }
      }
      break;
  }

}



void CGSEdit::OnChar( int iChar )
{

  if ( iChar == 'm' )
  {
    m_EditMode = (EditMode)( ( m_EditMode + 1 ) % EM_LAST_ENTRY );
  }
  else if ( iChar == 27 )
  {
    theApp.NextState( new CGSMenu() );
  }
  else if ( iChar == '1' )
  {
    m_EditWorld = "world";
    m_pSelectedObj = NULL;
    m_Level.Load( m_EditWorld );
  }
  else if ( iChar == '2' )
  {
    m_EditWorld = "dungeon";
    m_pSelectedObj = NULL;
    m_Level.Load( m_EditWorld );
  }

}



void CGSEdit::OnKeyDown( int iChar )
{

  if ( iChar == VK_DELETE )
  {
    if ( ( m_EditMode == EM_OBJECTS )
    &&   ( m_pSelectedObj != NULL ) )
    {
      m_Level.RemoveObject( m_pSelectedObj );
      delete m_pSelectedObj;
      m_pSelectedObj = NULL;
    }
  }

  if ( ( iChar == VK_ADD )
  ||   ( iChar == VK_OEM_PLUS ) )
  {
    if ( m_EditMode == EM_TILES )
    {
      m_CurTile = (LD::TileType)( ( m_CurTile + 1 ) % LD::TT_LAST_ENTRY );
    }
    else if ( m_EditMode == EM_OBJECTS )
    {
      m_CurObjectType = (LD::ObjectType)( ( m_CurObjectType + 1 ) % LD::OT_LAST_ENTRY );
    }
  }
  else if ( ( iChar == VK_SUBTRACT )
  ||        ( iChar == VK_OEM_MINUS ) )
  {
    if ( m_EditMode == EM_TILES )
    {
      m_CurTile = (LD::TileType)( ( m_CurTile + LD::TT_LAST_ENTRY - 1 ) % LD::TT_LAST_ENTRY );
    }
    else if ( m_EditMode == EM_OBJECTS )
    {
      m_CurObjectType = (LD::ObjectType)( ( m_CurObjectType + LD::OT_LAST_ENTRY - 1 ) % LD::OT_LAST_ENTRY );
    }
  }

}