#include "Level.h"
#include "LD.h"
#include "GSDone.h"
#include "Behaviour.h"
#include "BehaviourPlayer.h"

#include <set>
#include <string>
#include <fstream>
#include <vector>
#include <sstream>



Level::Sector::Sector()
{
  for ( int i = 0; i < 5; ++i )
  {
    for ( int j = 0; j < 5; ++j )
    {
      Ground[i][j] = LD::TT_GRASS;
    }
  }
}



void Level::Sector::Clear()
{
  std::list<GameObject*>::iterator      it( Objects.begin() );
  while ( it != Objects.end() )
  {
    delete *it;

    ++it;
  }
  Objects.clear();
  for ( int i = 0; i < 5; ++i )
  {
    for ( int j = 0; j < 5; ++j )
    {
      Ground[i][j] = LD::TT_GRASS;
    }
  }
}



Level::Level() :
  m_OffsetX( 0 ),
  m_OffsetY( 0 ),
  m_OffsetXFract( 0.0f ),
  m_OffsetYFract( 0.0f ),
  m_pPlayer( NULL ),
  m_DisplayTextTime( 0.0f ),
  m_pPlayerBehaviour( NULL ),
  m_Done( false ),
  m_DoneTime( 0 )
{
}



Level::~Level()
{

  Clear();

}



void Level::Clear()
{

  std::map<std::pair<int,int>,Sector>::iterator   it( m_Sectors.begin() );
  while ( it != m_Sectors.end() )
  {
    it->second.Clear();

    ++it;
  }
  m_Sectors.clear();
  m_ExtraData.clear();
  m_DisplayTextTime = 0.0f;

}



void Level::DisplayVisibleBounds()
{

  int     offsetX = m_OffsetX - 640 / 2;
  int     offsetY = m_OffsetY - 480 / 2;

  int     rightEnd = offsetX + 640;
  int     bottomEnd = offsetY + 480;

  int     secX1 = offsetX / 160;
  if ( offsetX < 0 )
  {
    --secX1;
  }
  int     secX2 = secX1 + 640 / 160;
  int     secY1 = offsetY / 160;
  if ( offsetY < 0 )
  {
    --secY1;
  }
  int     secY2 = secY1 + 480 / 160;

  theApp.m_pd3dDevice->SetTexture( 0, NULL );

  for ( int y = secY1; y <= secY2; ++y )
  {
    for ( int x = secX1; x <= secX2; ++x )
    {
      Level::Sector& sector( m_Sectors[std::make_pair( x, y )] );

      std::list<GameObject*>::iterator    itO( sector.Objects.begin() );
      while ( itO != sector.Objects.end() )
      {
        GameObject*   pObj( *itO );

        RECT    bounds = pObj->Bounds();

        OffsetRect( &bounds, -m_OffsetX + 320, -m_OffsetY + 240 );

        DWORD   color = 0xff00ff00;
        theApp.RenderQuad2d( bounds.left, bounds.top, bounds.right - bounds.left, 1, color );
        theApp.RenderQuad2d( bounds.left, bounds.top, 1, bounds.bottom - bounds.top, color );
        theApp.RenderQuad2d( bounds.right - 1, bounds.top, 1, bounds.bottom - bounds.top, color );
        theApp.RenderQuad2d( bounds.left, bounds.bottom - 1, bounds.right - bounds.left, 1, color );

        ++itO;
      }
    }
  }
}

    
    
void Level::DisplaySector( int X, int Y, Sector& Sector )
{

  int     secXOffset = -m_OffsetX + X * 5 * 32 + 320;
  int     secYOffset = -m_OffsetY + Y * 5 * 32 + 240;

  for ( int i = 0; i < 5; ++i )
  {
    for ( int j = 0; j < 5; ++j )
    {
      theApp.RenderTextureSectionAsTile2d( secXOffset + i * 32, secYOffset + j * 32, theApp.m_TexSec[theApp.m_Tile[Sector.Ground[i][j]]] );
    }
  }
  std::list<GameObject*>::iterator      itO( Sector.Objects.begin() );
  while ( itO != Sector.Objects.end() )
  {
    GameObject* pObj( *itO );

    m_ObjectsToDisplay.insert( std::make_pair( pObj->m_Y, pObj ) );

    ++itO;
  }

  /*
  char  temp[200];
  sprintf( temp, "%d,%d", X, Y );
  theApp.Print( secXOffset, secYOffset, temp );
  */

}



void Level::Display()
{

  int     offsetX = m_OffsetX - 640 / 2 - 64;
  int     offsetY = m_OffsetY - 480 / 2 - 64;

  int     secX1 = offsetX / 160;
  if ( offsetX < 0 )
  {
    --secX1;
  }
  int     secX2 = secX1 + ( 128 + 640 ) / 160 + 1;
  int     secY1 = offsetY / 160;
  if ( offsetY < 0 )
  {
    --secY1;
  }
  int     secY2 = secY1 + ( 128 + 480 ) / 160 + 1;

  m_ObjectsToDisplay.clear();

  for ( int y = secY1; y <= secY2; ++y )
  {
    for ( int x = secX1; x <= secX2; ++x )
    {
      DisplaySector( x, y, m_Sectors[std::make_pair( x, y )] );
    }
  }

  std::multimap<int,GameObject*>::iterator    itO( m_ObjectsToDisplay.begin() );
  while ( itO != m_ObjectsToDisplay.end() )
  {
    GameObject* pObj( itO->second );

    pObj->Display( *this );
    ++itO;
  }

  if ( m_DisplayTextTime > 0.0f )
  {
    theApp.PrintNiceCentered( 320 + m_DisplayTextPos.x - m_OffsetX, 240 + m_DisplayTextPos.y - m_OffsetY, m_DisplayText.c_str() );
  }

}



void Level::Update( float ElapsedTime )
{

  int     offsetX = m_OffsetX - 640 / 2 - 64;
  int     offsetY = m_OffsetY - 480 / 2 - 64;

  int     secX1 = offsetX / 160;
  if ( offsetX < 0 )
  {
    --secX1;
  }
  int     secX2 = secX1 + ( 128 + 640 ) / 160 + 1;
  int     secY1 = offsetY / 160;
  if ( offsetY < 0 )
  {
    --secY1;
  }
  int     secY2 = secY1 + ( 128 + 480 ) / 160 + 1;

  for ( int y = secY1; y <= secY2; ++y )
  {
    for ( int x = secX1; x <= secX2; ++x )
    {
      Sector& sector( m_Sectors[std::make_pair( x, y )] );

      std::list<GameObject*>::iterator      itO( sector.Objects.begin() );
      while ( itO != sector.Objects.end() )
      {
        GameObject* pObj( *itO );

        if ( pObj->m_RemoveMe )
        {
          delete pObj;
          itO = sector.Objects.erase( itO );
        }
        else
        {
          pObj->Update( ElapsedTime, *this );
          ++itO;
        }
      }
    }
  }

  while ( !m_DelayedEvents.empty() )
  {
    std::pair<std::string,int>  delayedEvent = m_DelayedEvents.front();
    m_DelayedEvents.pop_front();

    HandleExtraData( delayedEvent.first, delayedEvent.second );
  }

  if ( m_pPlayer )
  {
    m_OffsetX = m_pPlayer->m_X;
    m_OffsetY = m_pPlayer->m_Y;
  }

}



void Level::CheckCollisions( GameObject* pObj, int X, int Y )
{

  RECT    ourBounds = pObj->Bounds();
  RECT    dummyRect;

  for ( int y = Y - 1; y <= Y + 1; ++y )
  {
    for ( int x = X - 1; x <= X + 1; ++x )
    {
      std::map<std::pair<int,int>,Sector>::iterator   it( m_Sectors.find( std::make_pair( x, y ) ) );
      if ( it != m_Sectors.end() )
      {
        Sector&   sector( it->second );

        std::list<GameObject*>::iterator      itO( sector.Objects.begin() );
        while ( itO != sector.Objects.end() )
        {
          GameObject* pOtherObj( *itO );

          if ( ( pObj != pOtherObj )
          &&   ( !pOtherObj->m_RemoveMe )
          &&   ( ( CanCollide( pObj, pOtherObj ) )
          ||     ( CanCollide( pOtherObj, pObj ) ) )
          &&   ( IntersectRect( &dummyRect, &ourBounds, &pOtherObj->Bounds() ) ) )
          {
            OnCollision( pObj, pOtherObj );
            OnCollision( pOtherObj, pObj );
          }

          if ( pObj->m_RemoveMe )
          {
            return;
          }

          ++itO;
        }
      }
    }
  }

}



void Level::CheckVisibleCollisions()
{

  int     offsetX = m_OffsetX - 640 / 2 - 64;
  int     offsetY = m_OffsetY - 480 / 2 - 64;

  int     secX1 = offsetX / 160;
  if ( offsetX < 0 )
  {
    --secX1;
  }
  int     secX2 = secX1 + ( 128 + 640 ) / 160 + 1;
  int     secY1 = offsetY / 160;
  if ( offsetY < 0 )
  {
    --secY1;
  }
  int     secY2 = secY1 + ( 128 + 480 ) / 160 + 1;

  m_ObjectsToDisplay.clear();

  for ( int y = secY1; y <= secY2; ++y )
  {
    for ( int x = secX1; x <= secX2; ++x )
    {
      std::map<std::pair<int,int>,Sector>::iterator   it( m_Sectors.find( std::make_pair( x, y ) ) );
      if ( it != m_Sectors.end() )
      {
        Sector&   sector( it->second );

        std::list<GameObject*>::iterator      itO( sector.Objects.begin() );
        while ( itO != sector.Objects.end() )
        {
          GameObject* pObj( *itO );

          CheckCollisions( pObj, x, y );
          ++itO;
        }
      }
    }
  }

}



void Level::UpdateFixed()
{
  
  if ( m_Done )
  {
    m_DoneTime++;
    if ( m_DoneTime >= 300 )
    {
      theApp.NextState( new CGSDone() );
    }
  }

  int     offsetX = m_OffsetX - 640 / 2 - 64;
  int     offsetY = m_OffsetY - 480 / 2 - 64;

  int     secX1 = offsetX / 160;
  if ( offsetX < 0 )
  {
    --secX1;
  }
  int     secX2 = secX1 + ( 128 + 640 ) / 160 + 1;
  int     secY1 = offsetY / 160;
  if ( offsetY < 0 )
  {
    --secY1;
  }
  int     secY2 = secY1 + ( 128 + 480 ) / 160 + 1;

  for ( int y = secY1; y <= secY2; ++y )
  {
    for ( int x = secX1; x <= secX2; ++x )
    {
      Sector& sector( m_Sectors[std::make_pair( x, y )] );

      std::list<GameObject*>::iterator      itO( sector.Objects.begin() );
      while ( itO != sector.Objects.end() )
      {
        GameObject* pObj( *itO );

        if ( pObj->m_RemoveMe )
        {
          delete pObj;
          itO = sector.Objects.erase( itO );
        }
        else
        {
          ++itO;
          pObj->UpdateFixed( *this );
        }
      }
    }
  }

  CheckVisibleCollisions();

  if ( m_pPlayer )
  {
    m_OffsetX = m_pPlayer->m_X;
    m_OffsetY = m_pPlayer->m_Y;
  }

}



GameObject* Level::FindObjectByID( int ID )
{
  std::map<std::pair<int,int>,Sector>::iterator   it( m_Sectors.begin() );
  while ( it != m_Sectors.end() )
  {
    Sector& sector( it->second );

    std::list<GameObject*>::iterator      itO( sector.Objects.begin() );
    while ( itO != sector.Objects.end() )
    {
      GameObject* pObj( *itO );
      
      if ( pObj->m_ID == ID )
      {
        return pObj;
      }
      ++itO;
    }
    ++it;
  }
  return NULL;
}



bool Level::Save( const std::string& Filename )
{

  char    realFilename[200];
  sprintf( realFilename, "data/%s.dat", Filename.c_str() );

  FILE*   handle = fopen( realFilename, "wb" );
  if ( handle == NULL )
  {
    return false;
  }

  int   dummy = 0;

  dummy = (int)m_Sectors.size();
  fwrite( &dummy, sizeof( dummy ), 1, handle );

  std::map<std::pair<int,int>,Sector>::iterator   itS( m_Sectors.begin() );
  while ( itS != m_Sectors.end() )
  {
    dummy = 0x1000;
    fwrite( &dummy, sizeof( dummy ), 1, handle );

    fwrite( &itS->first.first, sizeof( itS->first.first ), 1, handle );
    fwrite( &itS->first.second, sizeof( itS->first.second ), 1, handle );

    // sector data
    Sector&   sector( itS->second );

    fwrite( sector.Ground, sizeof( sector.Ground ), 1, handle );

    // objects
    dummy = (int)sector.Objects.size();
    fwrite( &dummy, sizeof( dummy ), 1, handle );

    std::list<GameObject*>::iterator    itO( sector.Objects.begin() );
    while ( itO != sector.Objects.end() )
    {
      GameObject*   pObj( *itO );

      fwrite( &pObj->m_Type, sizeof( pObj->m_Type ), 1, handle );
      fwrite( &pObj->m_X, sizeof( pObj->m_X ), 1, handle );
      fwrite( &pObj->m_Y, sizeof( pObj->m_Y ), 1, handle );
      fwrite( &pObj->m_ID, sizeof( pObj->m_ID ), 1, handle );

      // reserve
      dummy = 0;
      fwrite( &dummy, sizeof( dummy ), 1, handle );
      fwrite( &dummy, sizeof( dummy ), 1, handle );

      ++itO;
    }

    ++itS;
  }

  fclose( handle );
  return true;

}



bool Level::Load( const std::string& Filename )
{

  Clear();

  m_Worldname = Filename;
  char    realFilename[200];
  sprintf( realFilename, "data/%s.dat", Filename.c_str() );

  FILE*   handle = fopen( realFilename, "rb" );
  if ( handle == NULL )
  {
    return false;
  }

  int   dummy = 0;
  int   numSectors = 0;

  fread( &numSectors, sizeof( numSectors ), 1, handle );
  
  for ( int i = 0; i < numSectors; ++i )
  {
    dummy = 0x1000;
    fread( &dummy, sizeof( dummy ), 1, handle );
    if ( dummy != 0x1000 )
    {
      return false;
    }

    int     sectorX = 0;
    int     sectorY = 0;

    fread( &sectorX, sizeof( sectorX ), 1, handle );
    fread( &sectorY, sizeof( sectorY ), 1, handle );

    // sector data
    Sector&   sector( m_Sectors[std::make_pair( sectorX, sectorY )] );

    fread( sector.Ground, sizeof( sector.Ground ), 1, handle );

    // objects
    int numObjects = 0;
    fread( &numObjects, sizeof( numObjects ), 1, handle );

    for ( int j = 0; j < numObjects; ++j )
    {
      LD::ObjectType    objType;
      int     X = 0;
      int     Y = 0;
      int     ID = 0;

      fread( &objType, sizeof( objType ), 1, handle );
      fread( &X, sizeof( X ), 1, handle );
      fread( &Y, sizeof( Y ), 1, handle );
      fread( &ID, sizeof( ID ), 1, handle );

      if ( m_SingleKillItems[m_Worldname].find( ID ) == m_SingleKillItems[m_Worldname].end() )
      {
        GameObject*   pObj = new GameObject( X, Y, objType );
        pObj->m_ID = ID;
        AddObject( pObj );

        if ( m_OpenItems[m_Worldname].find( ID ) != m_OpenItems[m_Worldname].end() )
        {
          pObj->OnEvent( *this, "Open", pObj, NULL );
        }
      }

      // reserve
      fread( &dummy, sizeof( dummy ), 1, handle );
      fread( &dummy, sizeof( dummy ), 1, handle );
    }
  }

  fclose( handle );

  char    temp2[200];
  sprintf( temp2, "%s.extra", realFilename );

  std::ifstream iFile( temp2 );        

  while ( ( iFile.good() )
  &&      ( !iFile.eof() ) )
  {
    std::string     line;

    std::getline( iFile, line, '\n' );
    //iFile >> line;
    if ( ( line.empty() )
    &&   ( iFile.eof() ) )
    {
      break;
    }

    size_t    sepPos = line.find( ':' );
    if ( sepPos != std::string::npos )
    {
      std::string     IDText = line.substr( 0, sepPos );
      std::string     Data = line.substr( sepPos + 1 );

      int     ID = atoi( IDText.c_str() );

      m_ExtraData[ID].push_back( Data );
    }
  }
  iFile.close();

  return true;

}



void Level::AddObject( GameObject* pObj )
{

  int   sectX = pObj->m_X / 160;
  if ( pObj->m_X < 0 )
  {
    --sectX;
  }
  int   sectY = pObj->m_Y / 160;
  if ( pObj->m_Y < 0 )
  {
    --sectY;
  }

  m_Sectors[std::make_pair( sectX, sectY )].Objects.push_back( pObj );

}



void Level::RemoveObject( GameObject* pObj )
{

  int   sectX = pObj->m_X / 160;
  if ( pObj->m_X < 0 )
  {
    --sectX;
  }
  int   sectY = pObj->m_Y / 160;
  if ( pObj->m_Y < 0 )
  {
    --sectY;
  }
  m_Sectors[std::make_pair( sectX, sectY )].Objects.remove( pObj );

}



void Level::MoveObject( GameObject* pObj, int NewX, int NewY )
{

  int   sectX = pObj->m_X / 160;
  if ( pObj->m_X < 0 )
  {
    --sectX;
  }
  int   sectY = pObj->m_Y / 160;
  if ( pObj->m_Y < 0 )
  {
    --sectY;
  }

  int   newSectX = NewX / 160;
  if ( NewX < 0 )
  {
    --newSectX;
  }
  int   newSectY = NewY / 160;
  if ( NewY < 0 )
  {
    --newSectY;
  }

  pObj->m_X = NewX;
  pObj->m_Y = NewY;
  if ( ( newSectX != sectX )
  ||   ( newSectY != sectY ) )
  {
    m_Sectors[std::make_pair( sectX, sectY )].Objects.remove( pObj );
    m_Sectors[std::make_pair( newSectX, newSectY )].Objects.push_back( pObj );
  }

}



int Level::TileXFromGlobalPos( int X )
{
  int     tileX = X / 32;
  if ( X < 0 )
  {
    tileX = ( X + 1 ) / 32 - 1;
    //tileX--;
  }
  return tileX;
}


  
int Level::TileXFromScreenPos( int X )
{
  return TileXFromGlobalPos( m_OffsetX + X );
}



int Level::TileYFromGlobalPos( int Y )
{
  int     tileY = Y / 32;
  if ( Y < 0 )
  {
    tileY = ( Y + 1 ) / 32 - 1;
    //tileY--;
  }
  return tileY;
}



int Level::TileYFromScreenPos( int Y )
{
  return TileYFromGlobalPos( m_OffsetY + Y );
}



//#include <debug/debugclient.h>
bool Level::AreaBlocked( GameObject* pMover, const RECT& rect, int DX, int DY )
{

  RECT    newRect = rect;

  //OffsetRect( &newRect, DX - 320, DY - 240 );
  OffsetRect( &newRect, DX, DY );

  int     tileX1 = TileXFromGlobalPos( newRect.left );
  int     tileX2 = TileXFromGlobalPos( newRect.right - 1 );
  int     tileY1 = TileYFromGlobalPos( newRect.top );
  int     tileY2 = TileYFromGlobalPos( newRect.bottom - 1 );

  int     sectX1 = tileX1 / 5;
  if ( tileX1 < 0 )
  {
    sectX1 = ( tileX1 + 1 ) / 5 - 1;
  }
  int   sectY1 = tileY1 / 5;
  if ( tileY1 < 0 )
  {
    sectY1 = ( tileY1 + 1 ) / 5 - 1;
  }
  int     sectX2 = tileX2 / 5;
  if ( tileX2 < 0 )
  {
    sectX2 = ( tileX2 + 1 ) / 5 - 1;
  }
  int   sectY2 = tileY2 / 5;
  if ( tileY2 < 0 )
  {
    sectY2 = ( tileY2 + 1 ) / 5 - 1;
  }

  //dh::Log( "---------------" );
  for ( int x = tileX1; x <= tileX2; ++x )
  {
    for ( int y = tileY1; y <= tileY2; ++y )
    {
      int   sectX = x / 5;
      if ( x < 0 )
      {
        sectX = ( x + 1 ) / 5 - 1;
      }
      int   sectY = y / 5;
      if ( y < 0 )
      {
        sectY = ( y + 1 ) / 5 - 1;
      }
      int     inSectX = x - sectX * 5;
      int     inSectY = y - sectY * 5;

      //dh::Log( "Check tile %d,%d, Sector %d,%d, InSector %d,%d", x, y, sectX, sectY, inSectX, inSectY );

      LD::TileType  tile = m_Sectors[std::make_pair( sectX, sectY )].Ground[inSectX][inSectY];

      if ( ( tile == LD::TT_ROCK )
      ||   ( tile == LD::TT_BRICK ) )
      {
        return true;
      }
      if ( ( tile == LD::TT_WATER )
      &&   ( m_pPlayerBehaviour )
      &&   ( m_pPlayerBehaviour->m_Items[LD::OT_WATER_SHOES] == 0 ) )
      {
        return true;
      }
    }
  }

  RECT    dummy;
  for ( int x = sectX1 - 1; x <= sectX2 + 1; ++x )
  {
    for ( int y = sectY1 - 1; y <= sectY2 + 1; ++y )
    {
      Level::Sector& sector( m_Sectors[std::make_pair( x, y )] );

      std::list<GameObject*>::iterator    itO( sector.Objects.begin() );
      while ( itO != sector.Objects.end() )
      {
        GameObject*   pObj( *itO );

        if ( IsBlocking( pMover->m_Type, pObj->m_Type ) )
        {
          RECT    objBounds = pObj->Bounds();

          if ( IntersectRect( &dummy, &newRect, &objBounds ) )
          {
            return true;
          }
        }

        ++itO;
      }

    }
  }
  return false;


}



int Level::WalkObject( GameObject* pObj, int DX, int DY )
{

  RECT    bounds = pObj->Bounds();

  int     result = 0;

  while ( ( DX != 0 )
  ||      ( DY != 0 ) )
  {
    if ( DX > 0 )
    {
      if ( AreaBlocked( pObj, bounds, 1, 0 ) )
      {
        DX = 0;
        result |= 1;
      }
      else
      {
        MoveObject( pObj, pObj->m_X + 1, pObj->m_Y );
        OffsetRect( &bounds, 1, 0 );
        --DX;
      }
    }
    else if ( DX < 0 )
    {
      if ( AreaBlocked( pObj, bounds, -1, 0 ) )
      {
        DX = 0;
        result |= 1;
      }
      else
      {
        MoveObject( pObj, pObj->m_X - 1, pObj->m_Y );
        OffsetRect( &bounds, -1, 0 );
        ++DX;
      }
    }
    if ( DY > 0 )
    {
      if ( AreaBlocked( pObj, bounds, 0, 1 ) )
      {
        DY = 0;
        result |= 2;
      }
      else
      {
        MoveObject( pObj, pObj->m_X, pObj->m_Y + 1 );
        OffsetRect( &bounds, 0, 1 );
        --DY;
      }
    }
    else if ( DY < 0 )
    {
      if ( AreaBlocked( pObj, bounds, 0, -1 ) )
      {
        DY = 0;
        result |= 2;
      }
      else
      {
        MoveObject( pObj, pObj->m_X, pObj->m_Y - 1 );
        OffsetRect( &bounds, 0, -1 );
        ++DY;
      }
    }
  }
  return result;

}



int Level::NewID()
{

  std::set<int>   usedIDs;

  std::map<std::pair<int,int>,Sector>::iterator   it( m_Sectors.begin() );
  while ( it != m_Sectors.end() )
  {
    Sector& sector( it->second );

    std::list<GameObject*>::iterator      itO( sector.Objects.begin() );
    while ( itO != sector.Objects.end() )
    {
      GameObject* pObj( *itO );

      usedIDs.insert( pObj->m_ID );

      ++itO;
    }
    ++it;
  }

  int   newID = 1;

  while ( usedIDs.find( newID ) != usedIDs.end() )
  {
    ++newID;
  }
  return newID;

}



GameObject* Level::FindObject( int X, int Y )
{

  int     offsetX = m_OffsetX - 640 / 2;
  int     offsetY = m_OffsetY - 480 / 2;

  int     rightEnd = offsetX + 640;
  int     bottomEnd = offsetY + 480;

  int     secX1 = offsetX / 160;
  if ( offsetX < 0 )
  {
    --secX1;
  }
  int     secX2 = secX1 + 640 / 160;
  int     secY1 = offsetY / 160;
  if ( offsetY < 0 )
  {
    --secY1;
  }
  int     secY2 = secY1 + 480 / 160;

  m_ObjectsToDisplay.clear();

  POINT   mousePos;

  mousePos.x = X + m_OffsetX;
  mousePos.y = Y + m_OffsetY;

  for ( int y = secY1; y <= secY2; ++y )
  {
    for ( int x = secX1; x <= secX2; ++x )
    {
      Sector&   sector( m_Sectors[std::make_pair( x, y )] );

      std::list<GameObject*>::iterator    itO( sector.Objects.begin() );
      while ( itO != sector.Objects.end() )
      {
        GameObject*   pObj( *itO );

        RECT    bounds = pObj->Bounds();

        if ( PtInRect( &bounds, mousePos ) )
        {
          return pObj;
        }
        ++itO;
      }
    }
  }
  return NULL;

}



bool Level::CanCollide( GameObject* pObj1, GameObject* pObj2 )
{

  if ( pObj1->m_Type == LD::OT_PLAYER )
  {
    if ( ( pObj2->m_Type == LD::OT_ENTRANCE )
    ||   ( pObj2->m_Type == LD::OT_SWORD_COLLECTABLE )
    ||   ( pObj2->m_Type == LD::OT_AXE_COLLECTABLE )
    ||   ( pObj2->m_Type == LD::OT_HAMMER_COLLECTABLE )
    ||   ( pObj2->m_Type == LD::OT_GOLDEN_SWORD_COLLECTABLE )
    ||   ( pObj2->m_Type == LD::OT_WATER_SHOES )
    ||   ( pObj2->m_Type == LD::OT_OCTA )
    ||   ( pObj2->m_Type == LD::OT_OCTO )
    ||   ( pObj2->m_Type == LD::OT_CUBE )
    ||   ( pObj2->m_Type == LD::OT_FIRE )
    ||   ( pObj2->m_Type == LD::OT_HEART_CONTAINER )
    ||   ( pObj2->m_Type == LD::OT_ENEMY_SHOT )
    ||   ( pObj2->m_Type == LD::OT_HEART_COLLECTABLE )
    ||   ( pObj2->m_Type == LD::OT_COIN ) )
    {
      return true;
    }
  }
  else if ( pObj1->m_Type == LD::OT_SWORD )
  {
    if ( ( pObj2->m_Type == LD::OT_OCTO )
    ||   ( pObj2->m_Type == LD::OT_OCTA )
    ||   ( pObj2->m_Type == LD::OT_GREEN_BUSH ) )
    {
      return true;
    }
  }
  else if ( pObj1->m_Type == LD::OT_GOLDEN_SWORD )
  {
    if ( ( pObj2->m_Type == LD::OT_OCTO )
    ||   ( pObj2->m_Type == LD::OT_OCTA )
    ||   ( pObj2->m_Type == LD::OT_CUBE )
    ||   ( pObj2->m_Type == LD::OT_TURRET )
    ||   ( pObj2->m_Type == LD::OT_GREEN_BUSH ) )
    {
      return true;
    }
  }
  else if ( pObj1->m_Type == LD::OT_AXE )
  {
    if ( ( pObj2->m_Type == LD::OT_BUSH )
    ||   ( pObj2->m_Type == LD::OT_OCTO )
    ||   ( pObj2->m_Type == LD::OT_OCTA )
    ||   ( pObj2->m_Type == LD::OT_CHEST ) )
    {
      return true;
    }
  }
  else if ( pObj1->m_Type == LD::OT_HAMMER )
  {
    if ( ( pObj2->m_Type == LD::OT_POST )
    ||   ( pObj2->m_Type == LD::OT_OCTO )
    ||   ( pObj2->m_Type == LD::OT_OCTA )
    ||   ( pObj2->m_Type == LD::OT_TURRET )
    ||   ( pObj2->m_Type == LD::OT_CHEST ) )
    {
      return true;
    }
  }
  return false;

}



void Split( const std::string& strSource, const char cSeparator, std::vector<std::string>& listParts )
{
  size_t      iPos = 0;

  size_t      iStartPos = 0;

  while ( ( iPos = strSource.find( cSeparator, iStartPos ) ) != std::string::npos )
  {
    listParts.push_back( strSource.substr( iStartPos, iPos - iStartPos ) );
    iStartPos = iPos + 1;
  }
  if ( ( strSource.length() >= iStartPos )
  &&   ( !strSource.empty() ) )
  {
    listParts.push_back( strSource.substr( iStartPos, strSource.length() - iStartPos ) );
  }
}



void Level::ParseExtraData( int ID )
{

  std::map<int,std::list<std::string> >::iterator   itED( m_ExtraData.find( ID ) );
  if ( itED == m_ExtraData.end() )
  {
    return;
  }

  std::list<std::string>::iterator    itS( itED->second.begin() );
  while ( itS != itED->second.end() )
  {
    m_DelayedEvents.push_back( std::make_pair( *itS, ID ) );

    ++itS;
  }

}



void Level::HandleExtraData( const std::string& ExtraData, int ID )
{

  std::vector<std::string>    params;
  
  Split( ExtraData, ',', params );

  if ( params.empty() )
  {
    return;
  }
  if ( params[0] == "collect" )
  {
    m_SingleKillItems[m_Worldname].insert( ID );
  }
  else if ( params[0] == "open" )
  {
    m_OpenItems[m_Worldname].insert( ID );
  }
  else if ( ( params[0] == "warp" )
  &&        ( params.size() == 4 ) )
  {
    std::string     newLevel = params[1];
    int             newX = atoi( params[2].c_str() );
    int             newY = atoi( params[3].c_str() );

    RemoveObject( m_pPlayer );

    Load( newLevel );
    m_pPlayer->m_X = newX;
    m_pPlayer->m_Y = newY;
    m_pPlayer->OnEvent( *this, "LevelChanged", m_pPlayer, NULL );
    AddObject( m_pPlayer );

    m_OffsetX = m_pPlayer->m_X;
    m_OffsetY = m_pPlayer->m_Y;
  }
  else if ( ( params[0] == "text" )
  &&        ( params.size() == 4 ) )
  {
    m_DisplayText = params[1];
    m_DisplayTextPos.x = atoi( params[2].c_str() );
    m_DisplayTextPos.y = atoi( params[3].c_str() );
    m_DisplayTextTime  = 5.0f;
  }
  else if ( params[0] == "done" )
  {
    m_Done = true;
    m_DoneTime = 0;
    m_pPlayer->m_TexSection = LD::TEX_PLAYER_D;
  }

}



bool Level::IsBlocking( LD::ObjectType Mover, LD::ObjectType Obstacle )
{
  if ( Mover == LD::OT_LEAVES )
  {
    return false;
  }
  if ( ( ( Mover == LD::OT_COIN )
  ||     ( Mover == LD::OT_HEART_COLLECTABLE ) )
  &&   ( Obstacle == LD::OT_CHEST ) )
  {
    return false;
  }
  if ( ( Obstacle == LD::OT_CHEST )
  ||   ( Obstacle == LD::OT_TREE )
  ||   ( Obstacle == LD::OT_POST )
  ||   ( Obstacle == LD::OT_BUSH )
  ||   ( Obstacle == LD::OT_GREEN_BUSH ) )
  {
    return true;
  }

  return false;
}



void Level::OnCollision( GameObject* pObj1, GameObject* pObj2 )
{

  switch ( pObj1->m_Type )
  {
    case LD::OT_PLAYER:
      if ( pObj2->m_Type == LD::OT_ENTRANCE )
      {
        ParseExtraData( pObj2->m_ID );
      }
      else if ( ( pObj2->m_Type == LD::OT_SWORD_COLLECTABLE )
      ||        ( pObj2->m_Type == LD::OT_COIN )
      ||        ( pObj2->m_Type == LD::OT_WATER_SHOES )
      ||        ( pObj2->m_Type == LD::OT_HEART_COLLECTABLE )
      ||        ( pObj2->m_Type == LD::OT_HAMMER_COLLECTABLE )
      ||        ( pObj2->m_Type == LD::OT_GOLDEN_SWORD_COLLECTABLE )
      ||        ( pObj2->m_Type == LD::OT_HEART_CONTAINER )
      ||        ( pObj2->m_Type == LD::OT_AXE_COLLECTABLE ) )
      {
        ParseExtraData( pObj2->m_ID );
        pObj1->OnEvent( *this, "Collect", pObj1, pObj2 );
        pObj2->m_RemoveMe = true;
      }
      else if ( ( pObj2->m_Type == LD::OT_OCTO ) 
      ||        ( pObj2->m_Type == LD::OT_ENEMY_SHOT )
      ||        ( pObj2->m_Type == LD::OT_FIRE )
      ||        ( pObj2->m_Type == LD::OT_CUBE )
      ||        ( pObj2->m_Type == LD::OT_OCTA ) )
      {
        pObj1->OnEvent( *this, "Hit", pObj1, pObj2 );
      }
      break;
    case LD::OT_SWORD:
      if ( ( pObj2->m_Type == LD::OT_OCTO )
      ||   ( pObj2->m_Type == LD::OT_OCTA )
      ||   ( pObj2->m_Type == LD::OT_CHEST )
      ||   ( pObj2->m_Type == LD::OT_GREEN_BUSH ) )
      {
        pObj2->OnEvent( *this, "Hit", pObj2, pObj1 );
      }
      break;
    case LD::OT_AXE:
      if ( ( pObj2->m_Type == LD::OT_OCTO )
      ||   ( pObj2->m_Type == LD::OT_OCTA )
      ||   ( pObj2->m_Type == LD::OT_BUSH )
      ||   ( pObj2->m_Type == LD::OT_CHEST ) )
      {
        pObj2->OnEvent( *this, "Hit", pObj2, pObj1 );
      }
      break;
    case LD::OT_HAMMER:
      if ( ( pObj2->m_Type == LD::OT_OCTO )
      ||   ( pObj2->m_Type == LD::OT_OCTA )
      ||   ( pObj2->m_Type == LD::OT_TURRET )
      ||   ( pObj2->m_Type == LD::OT_POST )
      ||   ( pObj2->m_Type == LD::OT_CHEST ) )
      {
        pObj2->OnEvent( *this, "Hit", pObj2, pObj1 );
      }
      break;
    case LD::OT_GOLDEN_SWORD:
      if ( ( pObj2->m_Type == LD::OT_OCTO )
      ||   ( pObj2->m_Type == LD::OT_OCTA )
      ||   ( pObj2->m_Type == LD::OT_CUBE )
      ||   ( pObj2->m_Type == LD::OT_CHEST )
      ||   ( pObj2->m_Type == LD::OT_GREEN_BUSH ) )
      {
        pObj2->OnEvent( *this, "Hit", pObj2, pObj1 );
      }
      break;
  }

}



